@@ -1,9 +1,16 @@
 import os, yaml, inspect, airflow
 from airflow import DAG
+from absql import Runner
+from functools import partial
 from gusty.errors import NonexistentDagDirError
 from gusty.parsing import parse, default_parsers
-from gusty.parsing.loaders import GustyYAMLLoader
+from gusty.parsing.loaders import (
+    generate_loader,
+    default_constructors,
+    handle_user_constructors,
+)
 from gusty.importing import airflow_version, get_operator
+from gusty.external_sensor import create_external_sensor, make_external_task_name
 
 ###########################
 ## Version Compatability ##
@@ -14,11 +21,23 @@ if airflow_version > 1:
 
 if airflow_version > 1:
     from airflow.operators.latest_only import LatestOnlyOperator
-    from airflow.sensors.external_task import ExternalTaskSensor
 else:
     from airflow.operators.latest_only_operator import LatestOnlyOperator
-    from airflow.sensors.external_task_sensor import ExternalTaskSensor
 
+#####################
+## Wait for Params ##
+#####################
+
+AVAILABLE_WAIT_FOR_PARAMS = [
+    "poke_interval",
+    "timeout",
+    "retries",
+    "mode",
+    "soft_fail",
+    "execution_delta",
+    "execution_date_fn",
+    "check_existence",
+]
 
 #########################
 ## Schematic Functions ##
@@ -83,7 +102,28 @@ def parse_external_dependencies(external_dependencies):
     """
     Document
     """
-    return dict(j for i in external_dependencies for j in i.items())
+    if isinstance(external_dependencies, dict):
+        deps = []
+        for dag_name, dag_details in external_dependencies.items():
+            for task_name in dag_details["tasks"]:
+                deps.append((dag_name, task_name))
+
+    else:
+        deps = [j for i in external_dependencies for j in i.items()]
+    return deps
+
+
+def parse_wait_for_overrides(external_dependencies):
+    if isinstance(external_dependencies, dict):
+        dags = external_dependencies.keys()
+        dag_wait_for_overrides = {}
+        for dag in dags:
+            wait_for_specs = external_dependencies[dag].copy()
+            wait_for_specs = {k: v for k, v in wait_for_specs.items() if k != "tasks"}
+            dag_wait_for_overrides.update({dag: wait_for_specs})
+        return dag_wait_for_overrides
+    else:
+        return {}
 
 
 #######################
@@ -101,7 +141,8 @@ def _get_operator_parameters(operator):
 
 def build_task(spec, level_id, schematic):
     """
-    Given a task specification ("spec"), locate the operator and instantiate the object with args from the spec.
+    Given a task specification ("spec"), locate the operator and
+    instantiate the object with args from the spec.
     """
     operator = get_operator(spec["operator"])
 
@@ -172,6 +213,13 @@ def build_structure(schematic, parent_id, name, metadata):
     return structure
 
 
+def flatten_nested_lists(x):
+    if isinstance(x, list):
+        return [a for i in x for a in flatten_nested_lists(i)]
+    else:
+        return [x]
+
+
 ##################
 ## GustyBuilder ##
 ##################
@@ -200,6 +248,20 @@ class GustyBuilder:
             ), "parse_hooks should be a dict of file extensions and handler functions for file_path."
             self.parsers.update(kwargs["parse_hooks"])
 
+        # Create dag_constructors
+        dag_constructors = default_constructors.copy()
+        if kwargs["dag_constructors"]:
+            dag_constructors.update(
+                handle_user_constructors(kwargs["dag_constructors"])
+            )
+
+        # Generate loader with dag_constructors
+        self.loader = generate_loader(dag_constructors)
+
+        # Generate runner with dag_constructors
+        runner_context = {k.strip("!"): v for k, v in dag_constructors.items()}
+        self.runner = Runner(**runner_context)
+
         self.schematic = create_schematic(dag_dir, self.parsers)
 
         # DAG defaults - everything that's not task_group_defaults or wait_for_defaults
@@ -218,24 +280,26 @@ class GustyBuilder:
         )
 
         # external dependency / wait_for_defaults
-        self.wait_for_defaults = {"poke_interval": 10, "timeout": 60, "retries": 60}
+        self.wait_for_defaults = {"poke_interval": 10, "timeout": 3600, "retries": 60}
         if "wait_for_defaults" in kwargs.keys():
             user_wait_for_defaults = {
                 k: v
                 for k, v in kwargs["wait_for_defaults"].items()
-                if k
-                in [
-                    "poke_interval",
-                    "timeout",
-                    "retries",
-                    "mode",
-                    "soft_fail",
-                    "execution_date_fn",
-                    "check_existence",
-                ]
+                if k in AVAILABLE_WAIT_FOR_PARAMS
+                or k
+                in inspect.signature(
+                    airflow.models.BaseOperator.__init__
+                ).parameters.keys()
             }
             self.wait_for_defaults.update(user_wait_for_defaults)
 
+        if "wait_for_class" in kwargs.keys():
+            self.create_external_sensor = partial(
+                create_external_sensor, sensor=kwargs["wait_for_class"]
+            )
+        else:
+            self.create_external_sensor = create_external_sensor
+
         # We will accept multiple levels only for Airflow v2 and up
         # This will keep the TaskGroup logic of the Levels class
         # Solely for Airflow v2 and beyond
@@ -263,7 +327,7 @@ class GustyBuilder:
         level_metadata_path = self.schematic[id]["metadata_path"]
         if os.path.exists(level_metadata_path or ""):
             with open(level_metadata_path) as inf:
-                level_metadata = yaml.load(inf, GustyYAMLLoader)
+                level_metadata = yaml.load(inf, self.loader)
 
             # special case - default_args provided in both metadata_defaults and level_metadata
             if (
@@ -280,6 +344,13 @@ class GustyBuilder:
                 # updated and resolved attached back to level_metadata
                 level_metadata.update({"default_args": metadata_default_args})
 
+            # special case - wait_for_defaults
+            if (
+                self.schematic[id]["parent_id"] is None
+                and "wait_for_defaults" in level_metadata.keys()
+            ):
+                self.wait_for_defaults.update(level_metadata["wait_for_defaults"])
+
         else:
             level_metadata = {}
         metadata_defaults.update(level_metadata)
@@ -311,10 +382,10 @@ class GustyBuilder:
             root_externals = metadata_default_dependencies["external_dependencies"]
             assert isinstance(
                 root_externals, list
-            ), """Root external dependencies set in create_dag must be a list of dicts following the pattern {"dag_id": "task_id"}"""
+            ), """Root external dependencies set in create_dag must be a list of dicts following the pattern {"dag_id": "task_id"}"""  # noqa
             assert all(
                 [isinstance(dep, dict) for dep in root_externals]
-            ), """Root external dependencies set in create_dag must be a list of dicts following the pattern {"dag_id": "task_id"}"""
+            ), """Root external dependencies set in create_dag must be a list of dicts following the pattern {"dag_id": "task_id"}"""  # noqa
             self.schematic[id].update({"external_dependencies": root_externals})
 
     def check_metadata(self, id):
@@ -354,7 +425,11 @@ class GustyBuilder:
         """
         level_metadata = self.schematic[id]["metadata"]
         level_spec_paths = self.schematic[id]["spec_paths"]
-        level_specs = [parse(spec_path, self.parsers) for spec_path in level_spec_paths]
+        level_specs = [
+            parse(spec_path, self.parsers, self.loader, self.runner)
+            for spec_path in level_spec_paths
+        ]
+        level_specs = flatten_nested_lists(level_specs)
         if airflow_version > 1:
             level_structure = self.schematic[id]["structure"]
             level_name = self.schematic[id]["name"]
@@ -391,10 +466,10 @@ class GustyBuilder:
     def create_level_dependencies(self, id):
         """
         For a given level id, identify what would be considered a valid set of dependencies within the DAG
-        for that level, and then set any specified dependencies upstream of that level. An example here would be
-        a DAG has two .yml jobs and one subfolder, which (the subfolder) is turned into a TaskGroup. That TaskGroup
-        can validly depend on the two .yml jobs, so long as either of those task_ids are defined within the dependencies
-        section of the TaskGroup's METADATA.yml
+        for that level, and then set any specified dependencies upstream of that level. An example here would
+        be a DAG has two .yml jobs and one subfolder, which (the subfolder) is turned into a TaskGroup. That
+        TaskGroup can validly depend on the two .yml jobs, so long as either of those task_ids are defined
+        within the dependencies section of the TaskGroup's METADATA.yml
         """
         level_structure = self.schematic[id]["structure"]
         level_dependencies = self.schematic[id]["dependencies"]
@@ -502,9 +577,10 @@ class GustyBuilder:
 
     def create_task_external_dependencies(self, id):
         """
-        As with create_task_dependencies, for a given level id, parse all of the external dependencies in a task's
-        spec, then create and add those "wait_for_" tasks upstream of a given task. Note the Builder class must keep
-        a record of all "wait_for_" tasks as to not recreate the same task twice.
+        As with create_task_dependencies, for a given level id, parse all of the
+        external dependencies in a task's spec, then create and add those "wait_for_"
+        tasks upstream of a given task. Note the Builder class must keep a record of all
+        "wait_for_" tasks as to not recreate the same task twice.
         """
         level_specs = self.schematic[id]["specs"]
         level_tasks = self.schematic[id]["tasks"]
@@ -515,34 +591,30 @@ class GustyBuilder:
                 if spec["task_id"] == task_id and "external_dependencies" in spec.keys()
             }
             if len(task_spec_external_dependencies) > 0:
-                task_external_dependencies = [
-                    external_dependency
-                    for external_dependency in task_spec_external_dependencies[task_id]
-                ]
                 task_external_dependencies = parse_external_dependencies(
-                    task_external_dependencies
+                    task_spec_external_dependencies[task_id]
                 )
-                for (
-                    external_dag_id,
-                    external_task_id,
-                ) in task_external_dependencies.items():
-                    wait_for_task_name = (
-                        "wait_for_DAG_{x}".format(x=external_dag_id)
-                        if external_task_id == "all"
-                        else "wait_for_{x}".format(x=external_task_id)
+                wait_for_overrides = parse_wait_for_overrides(
+                    task_spec_external_dependencies[task_id]
+                )
+                for dag_task_pair in task_external_dependencies:
+                    (external_dag_id, external_task_id) = dag_task_pair
+
+                    wait_for_task_name = make_external_task_name(
+                        external_dag_id, external_task_id
                     )
+
                     if wait_for_task_name in self.wait_for_tasks.keys():
                         wait_for_task = self.wait_for_tasks[wait_for_task_name]
                         task.set_upstream(wait_for_task)
                     else:
-                        wait_for_task = ExternalTaskSensor(
-                            dag=get_top_level_dag(self.schematic),
-                            task_id=wait_for_task_name,
-                            external_dag_id=external_dag_id,
-                            external_task_id=(
-                                external_task_id if external_task_id != "all" else None
-                            ),
-                            **self.wait_for_defaults
+                        wait_for_task = self.create_external_sensor(
+                            get_top_level_dag(self.schematic),
+                            external_dag_id,
+                            external_task_id,
+                            self.wait_for_defaults,
+                            wait_for_overrides,
+                            wait_for_task_name,
                         )
                         self.wait_for_tasks.update({wait_for_task_name: wait_for_task})
                         task.set_upstream(wait_for_task)
@@ -556,30 +628,30 @@ class GustyBuilder:
         level_parent_id = self.schematic[id]["parent_id"]
         if level_parent_id is not None:
             if len(level_external_dependencies) > 0:
-                level_external_dependencies = parse_external_dependencies(
+                level_external_dependencies_parsed = parse_external_dependencies(
                     level_external_dependencies
                 )
-                for (
-                    external_dag_id,
-                    external_task_id,
-                ) in level_external_dependencies.items():
-                    wait_for_task_name = (
-                        "wait_for_DAG_{x}".format(x=external_dag_id)
-                        if external_task_id == "all"
-                        else "wait_for_{x}".format(x=external_task_id)
+                wait_for_overrides = parse_wait_for_overrides(
+                    level_external_dependencies
+                )
+                for dag_task_pair in level_external_dependencies_parsed:
+                    (external_dag_id, external_task_id) = dag_task_pair
+
+                    wait_for_task_name = make_external_task_name(
+                        external_dag_id, external_task_id
                     )
+
                     if wait_for_task_name in self.wait_for_tasks.keys():
                         wait_for_task = self.wait_for_tasks[wait_for_task_name]
                         level_structure.set_upstream(wait_for_task)
                     else:
-                        wait_for_task = ExternalTaskSensor(
-                            dag=get_top_level_dag(self.schematic),
-                            task_id=wait_for_task_name,
-                            external_dag_id=external_dag_id,
-                            external_task_id=(
-                                external_task_id if external_task_id != "all" else None
-                            ),
-                            **self.wait_for_defaults
+                        wait_for_task = self.create_external_sensor(
+                            get_top_level_dag(self.schematic),
+                            external_dag_id,
+                            external_task_id,
+                            self.wait_for_defaults,
+                            wait_for_overrides,
+                            wait_for_task_name,
                         )
                         self.wait_for_tasks.update({wait_for_task_name: wait_for_task})
                         level_structure.set_upstream(wait_for_task)
@@ -651,14 +723,14 @@ class GustyBuilder:
                 for task_id, task in valid_root_tasks.items():
                     assert (
                         len(task.upstream_task_ids) == 0
-                    ), "Task {id} in DAG {dag} is listed as a root task, but has dependencies listed: {dep_list}".format(
+                    ), "Task {id} in DAG {dag} is listed as a root task, but has dependencies listed: {dep_list}".format(  # noqa
                         id=task_id,
                         dag=self.schematic[id]["structure"]._dag_id,
                         dep_list=", ".join(task.upstream_task_ids),
                     )
                     assert (
                         len(task.downstream_task_ids) == 0
-                    ), "Task {id} in DAG {dag} is listed as a root task, but other tasks explicity depend on it: {dep_list}".format(
+                    ), "Task {id} in DAG {dag} is listed as a root task, but other tasks explicity depend on it: {dep_list}".format(  # noqa
                         id=task_id,
                         dag=self.schematic[id]["structure"]._dag_id,
                         dep_list=", ".join(task.downstream_task_ids),
@@ -681,33 +753,34 @@ class GustyBuilder:
             # Set root-level external dependencies
             if len(level_external_dependencies) > 0:
                 root_external_dependencies = {}
-                level_external_dependencies = parse_external_dependencies(
+                level_external_dependencies_parsed = parse_external_dependencies(
                     level_external_dependencies
                 )
-                for (
-                    external_dag_id,
-                    external_task_id,
-                ) in level_external_dependencies.items():
-                    wait_for_task_name = (
-                        "wait_for_DAG_{x}".format(x=external_dag_id)
-                        if external_task_id == "all"
-                        else "wait_for_{x}".format(x=external_task_id)
+                wait_for_overrides = parse_wait_for_overrides(
+                    level_external_dependencies
+                )
+                for dag_task_pair in level_external_dependencies_parsed:
+                    (external_dag_id, external_task_id) = dag_task_pair
+
+                    wait_for_task_name = make_external_task_name(
+                        external_dag_id, external_task_id
                     )
+
                     if wait_for_task_name in self.wait_for_tasks.keys():
                         wait_for_task = self.wait_for_tasks[wait_for_task_name]
                         root_external_dependencies.update(
                             {wait_for_task_name: wait_for_task}
                         )
                     else:
-                        wait_for_task = ExternalTaskSensor(
-                            dag=get_top_level_dag(self.schematic),
-                            task_id=wait_for_task_name,
-                            external_dag_id=external_dag_id,
-                            external_task_id=(
-                                external_task_id if external_task_id != "all" else None
-                            ),
-                            **self.wait_for_defaults
+                        wait_for_task = self.create_external_sensor(
+                            get_top_level_dag(self.schematic),
+                            external_dag_id,
+                            external_task_id,
+                            self.wait_for_defaults,
+                            wait_for_overrides,
+                            wait_for_task_name,
                         )
+
                         self.wait_for_tasks.update({wait_for_task_name: wait_for_task})
                         root_external_dependencies.update(
                             {wait_for_task_name: wait_for_task}
@@ -773,14 +846,14 @@ class GustyBuilder:
                 for task_id, task in valid_leaf_tasks.items():
                     assert (
                         len(task.upstream_task_ids) == 0
-                    ), "Task {id} in DAG {dag} is listed as a leaf task, but has dependencies listed: {dep_list}".format(
+                    ), "Task {id} in DAG {dag} is listed as a leaf task, but has dependencies listed: {dep_list}".format(  # noqa
                         id=task_id,
                         dag=self.schematic[id]["structure"]._dag_id,
                         dep_list=", ".join(task.upstream_task_ids),
                     )
                     assert (
                         len(task.downstream_task_ids) == 0
-                    ), "Task {id} in DAG {dag} is listed as a leaf task, but other tasks explicity depend on it: {dep_list}".format(
+                    ), "Task {id} in DAG {dag} is listed as a leaf task, but other tasks explicity depend on it: {dep_list}".format(  # noqa
                         id=task_id,
                         dag=self.schematic[id]["structure"]._dag_id,
                         dep_list=", ".join(task.downstream_task_ids),
