@@ -2,8 +2,8 @@ import pytest
 import os
 from airflow import DAG
 from airflow.models import BaseOperator
-from datetime import datetime, timedelta
-from airflow.utils.dates import days_ago
+from datetime import timedelta
+from gusty.utils import days_ago
 from gusty.building import create_schematic
 from gusty.parsing import default_parsers
 from gusty import create_dag
@@ -34,6 +34,12 @@ def dag_files(no_metadata_dir):
 
 @pytest.fixture(scope="session")
 def dag(no_metadata_dir):
+    def simple_constructor(hello=True):
+        if hello:
+            return "hello"
+        else:
+            return "goodbye"
+
     dag = create_dag(
         no_metadata_dir,
         description="A dag created without metadata",
@@ -49,6 +55,7 @@ def dag(no_metadata_dir):
             "retry_delay": timedelta(minutes=5),
         },
         external_dependencies=[{"external_dag": "a_root_level_external"}],
+        dag_constructors=[simple_constructor],
     )
     return dag
 
@@ -91,15 +98,26 @@ def test_task_overrides_defaults(dag):
 
 def test_top_level_depends_on_second_level(dag):
     top_level_task = dag.task_dict["top_level_task"]
-    top_level_task_dependencies = top_level_task.__dict__["_upstream_task_ids"]
+    top_level_task_dependencies = top_level_task.__dict__["upstream_task_ids"]
     assert "dependable_task" in top_level_task_dependencies
 
 
 def test_external_dependencies(dag):
     dependable_task = dag.task_dict["dependable_task"]
-    dependable_task_dependencies = dependable_task.__dict__["_upstream_task_ids"]
+    dependable_task_dependencies = dependable_task.__dict__["upstream_task_ids"]
     assert "wait_for_DAG_a_whole_dag" in dependable_task_dependencies
     assert "wait_for_external_task" in dependable_task_dependencies
+    assert "wait_for_external_task_2" in dependable_task_dependencies
+
+
+def test_external_dependencies_dict_tasks(dag):
+    ext_dep_dict_spec_task = dag.task_dict["task_overrides_retries"]
+    ext_dep_dict_spec_task_dependencies = ext_dep_dict_spec_task.__dict__[
+        "upstream_task_ids"
+    ]
+    assert "wait_for_DAG_a_whole_dag" in ext_dep_dict_spec_task_dependencies
+    assert "wait_for_another_task_1" in ext_dep_dict_spec_task_dependencies
+    assert "wait_for_another_task_2" in ext_dep_dict_spec_task_dependencies
 
 
 def test_tasks_created(dag, dag_files):
@@ -112,7 +130,6 @@ def test_tasks_created(dag, dag_files):
 
     expected_tasks = [*map(os.path.basename, dag_files)]
     expected_tasks = [*map(replace_extension, expected_tasks)]
-    tasks_present = [task in dag.task_dict.keys() for task in expected_tasks]
     tasks_are_operators = [
         isinstance(dag.task_dict[task], BaseOperator) for task in expected_tasks
     ]
@@ -132,7 +149,7 @@ def test_sensor_support(dag):
 def test_root_external_dependencies_accepted(dag):
     assert (
         "wait_for_a_root_level_external"
-        in dag.task_dict["latest_only"].__dict__["_downstream_task_ids"]
+        in dag.task_dict["latest_only"].__dict__["downstream_task_ids"]
     )
 
 
@@ -140,8 +157,27 @@ def test_root_external_dependencies_latest_only_order(dag):
     # Bad test name, but bascially nothing should be upstream of latest only,
     # And only the DAG-level external dependency is after latest_only
     latest_only = dag.task_dict["latest_only"].__dict__
-    latest_only_downstream_tasks = latest_only["_downstream_task_ids"]
-    latest_only_upstream_tasks = latest_only["_upstream_task_ids"]
+    latest_only_downstream_tasks = latest_only["downstream_task_ids"]
+    latest_only_upstream_tasks = latest_only["upstream_task_ids"]
     assert (
         len(latest_only_upstream_tasks) == 0 and len(latest_only_downstream_tasks) == 1
     )
+
+
+def test_constructors(dag):
+    assert dag.task_dict["sensor_task"].__dict__["email"] == "goodbye"
+    assert dag.task_dict["sensor_task"].__dict__["owner"] == "hello"
+    # Below is generated a default func available in absql
+    assert dag.task_dict["sensor_task"].__dict__["doc"] == "default_absql_func"
+
+
+def test_context_rendering_omits_sql(dag):
+    sql_task = dag.task_dict["sql_task"]
+    assert sql_task.__dict__["doc"] == "hello"
+    assert (
+        sql_task.__dict__["sql"] == "SELECT date FROM my_table WHERE date = {{ date }}"
+    )
+
+
+def test_context_rendering_py(dag):
+    assert dag.task_dict["py_task"].__dict__["python_callable"]() == "hey"
